/*****************************************************************************/

/*
 *      dvdpc.c -- An add-on to handle packet command level ioctls on IDE 
 *                 DVD drives as a module. This is a very thin layer on 
 * 	           top of the Linux cdrom driver.
 *
 *      Copyright (C) 1999-2000  Sigma Designs
 *                    written by Emmanuel Michon <emmanuel_michon@sdesigns.com>
 *      Copyright (c) 1996, 1997 David A. van Leeuwen
 *      Copyright (c) 1997, 1998 Erik Andersen <andersee@debian.org>
 *      Copyright (c) 1998, 1999 Jens Axboe
 *
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*****************************************************************************/

#include <linux/version.h>

#include <linux/config.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/delay.h>
#include <linux/timer.h>
#include <linux/slab.h>
#include <linux/interrupt.h>
#include <linux/errno.h>
#include <linux/cdrom.h>
#include <linux/ide.h>
#include <linux/completion.h>
#include <linux/mutex.h>

#include <scsi/scsi.h>	/* For SCSI -> ATAPI command conversion */

#include <asm/irq.h>
#include <asm/io.h>
#include <asm/byteorder.h>
#include <asm/uaccess.h>
#include <asm/unaligned.h>


//#include <linux/cdrom.h>
//#include <linux/ide.h>
/* VERBOSE_IDE_CD_ERRORS dumps the spec. const tables in object files.
   Since both dvdpc.c and ide-cd.c include ide-cd.h,
   there is a conflict in final link */
#define VERBOSE_IDE_CD_ERRORS 0
#include "ide-cd.h"

#include "common.h"
#include "dvdpc.h"
#include "dvdpcdev.h"

#define RXTXSIZE 33000 // actual requests are by 32k blocks

static char rxtxbuffer[RXTXSIZE];

int pc_ioctl(struct inode *, struct file *, unsigned int, unsigned long);

static struct file_operations pc_fops =
{
	.ioctl = pc_ioctl,
};


// this is the old req struct from kernel 2.2.13
struct dvdpc_atapi_request_sense {
#if defined(__BIG_ENDIAN_BITFIELD)
	unsigned char valid      : 1;
	unsigned char error_code : 7;
#elif defined(__LITTLE_ENDIAN_BITFIELD)
	unsigned char error_code : 7;
	unsigned char valid      : 1;
#else
#error "Please fix <asm/byteorder.h>"
#endif
	byte reserved1;
#if defined(__BIG_ENDIAN_BITFIELD)
	unsigned char reserved3  : 2;
	unsigned char ili        : 1;
	unsigned char reserved2  : 1;
	unsigned char sense_key  : 4;
#elif defined(__LITTLE_ENDIAN_BITFIELD)
	unsigned char sense_key  : 4;
	unsigned char reserved2  : 1;
	unsigned char ili        : 1;
	unsigned char reserved3  : 2;
#else
#error "Please fix <asm/byteorder.h>"
#endif
	byte info[4];
	byte sense_len;
	byte command_info[4];
	byte asc;
	byte ascq;
	byte fru;
	byte sense_key_specific[3];
};

// in kernel 2.6 this has changed - so we use our own function
static inline int dvdpc_verify_area(int type, const void __user * addr, unsigned long size)
{
        return access_ok(type,addr,size) ? 0 : -EFAULT;
}

// methods from ide-cd.c from 2.6
static void cdrom_saw_media_change (ide_drive_t *drive)
{
	struct cdrom_info *info = drive->driver_data;
	
	CDROM_STATE_FLAGS (drive)->media_changed = 1;
	CDROM_STATE_FLAGS (drive)->toc_valid = 0;
	info->nsectors_buffered = 0;
}


static void cdrom_prepare_request(ide_drive_t *drive, struct request *rq)
{
	struct cdrom_info *cd = drive->driver_data;

	ide_init_drive_cmd(rq);
	rq->cmd_type = REQ_TYPE_ATA_PC;
	rq->rq_disk = cd->disk;
}


static int cdrom_queue_packet_command(ide_drive_t *drive, struct request *rq)
{
	struct request_sense sense;
	int retries = 10;
	unsigned int flags = rq->cmd_flags;

	if (rq->sense == NULL)
		rq->sense = &sense;

	/* Start of retry loop. */
	do {
		int error;
		unsigned long time = jiffies;
		rq->cmd_flags = flags;

		error = ide_do_drive_cmd(drive, rq, ide_wait);
		time = jiffies - time;

		/* FIXME: we should probably abort/retry or something 
		 * in case of failure */
		if (rq->cmd_flags & REQ_FAILED) {
			/* The request failed.  Retry if it was due to a unit
			   attention status
			   (usually means media was changed). */
			struct request_sense *reqbuf = rq->sense;

			if (reqbuf->sense_key == UNIT_ATTENTION)
				cdrom_saw_media_change(drive);
			else if (reqbuf->sense_key == NOT_READY &&
				 reqbuf->asc == 4 && reqbuf->ascq != 4) {
				/* The drive is in the process of loading
				   a disk.  Retry, but wait a little to give
				   the drive time to complete the load. */
				ssleep(2);
			} else {
				/* Otherwise, don't retry. */
				retries = 0;
			}
			--retries;
		}

		/* End of retry loop. */
	} while ((rq->cmd_flags & REQ_FAILED) && retries >= 0);

	/* Return an error if the command failed. */
	return (rq->cmd_flags & REQ_FAILED) ? -EIO : 0;
}

// ioctl function to realize the 0x1234 packetcommand ioctl
int pc_ioctl(struct inode *i_node, struct file *filp,
		 unsigned int cmd, unsigned long arg)
{
	ide_drive_t *drive;
	int idx = MINOR(i_node->i_rdev)&DVDPC_MINOR_DEV_MASK;
	dvdpc_dev_t *dev = &dvdpc_dev_tab[idx];

	drive = dev->handle;
	
	switch (cmd) {
		
		case IOCTL_PACKET_COMMAND:
		{
			int stat;
			struct request req;
			struct request_sense reqsense;

			int len, lena;
			
			cdrom_prepare_request(drive, &req);

			req.sense = &reqsense;

			stat = dvdpc_verify_area (VERIFY_READ, (void *) arg, CDROM_PACKET_SIZE);
			if (stat)
				return stat;
			copy_from_user (&req.cmd, (void *) arg, CDROM_PACKET_SIZE);
			arg += CDROM_PACKET_SIZE;

			// neccessary??? - seems to be ok?!
			req.cmd_len = CDROM_PACKET_SIZE;

			stat = dvdpc_verify_area (VERIFY_READ, (void *) arg, sizeof (len));
			if (stat)
				return stat;
			copy_from_user (&len, (void *) arg , sizeof (len));
			arg += sizeof (len);
			
			// len can be negative, then meaning that we want to send something
			lena = len;
			if (lena  < 0) lena = -lena;
			
			if (len > 0)
			{
				stat = dvdpc_verify_area (VERIFY_WRITE,(void *) arg, lena);
				if (stat)
					return stat;
			}
			else if (len < 0)
			{
				stat = dvdpc_verify_area (VERIFY_READ,(void *) arg, lena);
				if (stat)
					return stat;
				copy_from_user (rxtxbuffer, (void*)arg, lena);
			}
			
			if (len != 0)
			{
				req.data = rxtxbuffer;
				req.data_len = lena;
			}

			stat = cdrom_queue_packet_command (drive, &req);
			
			if (len > 0)
				copy_to_user ((void *)arg, rxtxbuffer, lena);
			
			// whatever happens return sense data
			arg+=lena;
			copy_to_user((void *)arg,req.sense,sizeof(struct dvdpc_atapi_request_sense));

			// this return value is important
			return stat;
		}
		
		default:
			return -EINVAL;
	}
	
	return 0; // never reached anyways
}

int atapidvd_hook(void)
{
	unsigned int index,unit;

	dvdpc_dev_idx=0;
	
	// this code comes from ide_scan in ide.c
	for (index = 0; index < MAX_HWIFS; ++index) {
		ide_hwif_t *hwif = &ide_hwifs[index];
		if (!hwif->present)
			continue;
		
		for (unit = 0; unit < MAX_DRIVES; ++unit) {
			ide_drive_t *drive = &hwif->drives[unit];
			dvdpc_dev_t *dev;
			
			LOGINFO(1,"scanning device #%d on ribbon #%d",unit,index);
			
			ASSERT(DEBUG,(dvdpc_dev_idx<DVDPC_MAX_DEV),
			       -EINVAL,RAS);
			
			dev=&dvdpc_dev_tab[dvdpc_dev_idx];
			
			dev->fops = &pc_fops;
			
			// the following line is the only important line
			dev->handle = drive;
			
			dev->opentimes = 0;
			
			LOGINFO(1,"found IDE device #%d @%p",
				dvdpc_dev_idx,
				drive);
			
			dvdpc_dev_idx++;
		}
	}
	
	ASSERT(1,dvdpc_dev_idx,
	       -EINVAL,"found %d IDE devices total",dvdpc_dev_idx);
	
	return 0;
}
